/* This is my very first Arduino code, so pardon the quality.
 * It reads data from Wii nunchuck and transforms it into commands 
 * for Pleo robotic dino toy. For more details about the project
 * please visit http://robostuff.com/ which is worth visiting 
 * regardless this hack.
 */

#include <Wire.h>
#include <string.h>

#undef int
#include <stdio.h>

#define BAUD_RATE 9600 // 19200 for programming, 115200 for working with Pleo, 9600 for transmission via xbee

#define LED_PIN 13 // the pin the LED is attached to
#define GROUND_PIN PC2
#define POWER_PIN PC3
#define POWER_ON_LED_PIN 13
#define RECEIVE_LED_PIN 12

#define DELTA 20 // the joystick position tolerance

#define ACTION "motion play roar" // playes on Z anc C buttons pressed
#define J_TOP_LEFT "motion play com_walk_fl_v1"
#define J_TOP "motion play com_walk_fs_v1"
#define J_TOP_RIGHT "motion play com_walk_fr_v1"
#define J_LEFT "motion play hun_sniff_l"            //???
#define J_RIGHT "motion play hun_sniff_r"           //???
#define J_BOTTOM_LEFT "motion play com_walk_bl_v1"
#define J_BOTTOM "motion play hun_exit"          //???
#define J_BOTTOM_RIGHT "motion play com_walk_br_v1"

#define CMD_DISABLE_LOG "log disable all"

#define MIN_X 316
#define MAX_X 730
#define MIN_Y 380
#define MAX_Y 630

uint8_t buf[6];  // global buffer for storing nunchuck data, bad, bad ...

// setup everything
void setup()
{
  pinMode( POWER_ON_LED_PIN, OUTPUT );
  pinMode( RECEIVE_LED_PIN, OUTPUT );
  digitalWrite( POWER_ON_LED_PIN, HIGH );
  delay( 100 );           // wait for things to stabilize
  Serial.begin( BAUD_RATE );   // configure serial output
  wait_for_pleo();         // wait for Pleo to boot up
  Wire.begin();            // join i2c bus with address 0x52
  power_on( PC3, PC2 );    // poweron the nunchuck
  init_nunchuck();         // send the initilization handshake
}

// waits for pleo to boot. It is important that XBee coordinator connected to arduino is turned on first
// then the XBee node connected to Pleo should be turned on. 
static void wait_for_pleo()
{
  digitalWrite( RECEIVE_LED_PIN, HIGH );
  int wait = 1;
  int count = 0;
  int read = 0;
  while( wait )
  {
    if( Serial.available() )
    {
      read = Serial.read();
      if( read == '>' )
      {
          wait = 0;
      }
    }
    else
    {
      delay( 500 );
      Serial.print( "\r\n" );
    }
  }
  command( CMD_DISABLE_LOG, 500 ); // disable Pleo logging otherwise serial gets trashed and everything stops working in a minute
  digitalWrite( RECEIVE_LED_PIN, LOW );
}

void delay_read( int sleep )
{
  int read;
  while( Serial.available() )
  {
    read = Serial.read();
  }
  delay( sleep );
}

// turns on power for the nunchuck
static void power_on( byte pwr, byte gnd )
{
  DDRC |= _BV(pwr) | _BV(gnd);  // make outputs
  PORTC &=~ _BV(gnd);
  PORTC |=  _BV(pwr);
  delay(100);  // wait for things to stabilize
}

// initialize nunchuck
void init_nunchuck()
{
  Wire.beginTransmission( 0x52 ); // transmit to device 0x52
  Wire.send( 0x40 );              // sends memory address
  Wire.send( 0x00 );              // sends memory address
  Wire.endTransmission();       // stop transmitting
}

// sending zero to nunchuck makes it send its current vital data back
void send_zero()
{
  Wire.beginTransmission( 0x52 ); // transmit to device 0x52
  Wire.send( 0x00 );		 // sends one byte
  Wire.endTransmission();	 // stop transmitting
}

void loop()
{
  digitalWrite( RECEIVE_LED_PIN, LOW );
  Serial.print( "\r\n" );         // I don't know why without this line the loop is just not working.
  int cnt = 0;                  // number of bytes read from input
  Wire.requestFrom( 0x52, 6 );  // request 6 bytes from nunchuck
  while( Wire.available() )
  {
    buf[cnt++] = decrypt( Wire.receive() );	// receive byte as an integer
  }
  if( cnt >= 5 )
  {
    process_buffer();          // does the job
  }
  send_zero(); // send the request for next bytes
  digitalWrite( RECEIVE_LED_PIN, HIGH );
}

// positions head according to the joystick position
void position_head( int x, int y )
{
  int nh = ( ( x - 320 ) * 13 / 41 ) - 65;
  nh = nh < -65 ? -65 : nh;
  nh = nh > 65 ? 65 : nh;
  int nv = ( ( 380 - y ) * 15 / 25 ) + 75;
  nv = nv < -75 ? -75 : nv;
  nv = nv > 75 ? 75 : nv;
  Serial.print( "joint move 11 " );
  Serial.print( nh );
  Serial.print( "\r\n" );
  Serial.print( "joint move 12 " );
  Serial.print( nv );
  Serial.print( "\r\n" );
  delay_read( 1000 );
}

// positions tail according to the joystick position
void position_tail( int x, int y )
{
  int nh = ( ( x - 320 ) * 18 / 41 ) - 90;
  nh = nh < -90 ? -90 : nh;
  nh = nh > 90 ? 90 : nh;
  int nv = ( ( 380 - y ) * 18 / 25 ) + 90;
  nv = nv < -90 ? -90 : nv;
  nv = nv > 90 ? 90 : nv;
  Serial.print( "joint move 9 " );
  Serial.print( nh );
  Serial.print( "\r\n" );
  Serial.print( "joint move 10 " );
  Serial.print( nv );
  Serial.print( "\r\n" );
  delay_read( 1000 );
}

void command( char cmd[], int sleep )
{
  Serial.print( cmd );
  Serial.print( "\r\n" );
  delay_read( sleep );
}

// parses return from nunchuck (6 bytes)
void process_buffer()
{
  int jx = buf[0];
  int jy = buf[1];
  int ax = buf[2] << 2 | ( (buf[5] >> 2 ) & 0x03 ); 
  int ay = buf[3] << 2 | ( (buf[5] >> 4 ) & 0x03 );
  int az = buf[4] << 2 | ( (buf[5] >> 6 ) & 0x03 );
  int bz = buf[5] & 1;
  int bc = ( buf[5] >> 1 )& 1;

  if( check( jx, jy, 53, 205 ) )
  {
    command( J_TOP_LEFT, 2000 );
  } 
  else if( check( jx, jy, 123, 226 ) )
  {
    command( J_TOP, 2000 );
  }
  else if( check( jx, jy, 202, 202 ) )
  {
    command( J_TOP_RIGHT, 2000 );
  }
  else if( check( jx, jy, 25, 127 ) )
  {
    command( J_LEFT, 2000 );
  }
  else if( check( jx, jy, 128, 132 ) )
  {
    if( ( bz == 1 ) && ( bc == 1 ) )
    {
      position_head( ax, ay );
    }
    else if( ( bz == 0 ) && ( bc == 1 ) )
    {
      position_tail( ax, ay );
    }
    else if( ( bz == 0 ) && ( bc == 0 ) )
    {
      command( ACTION, 4000 );
    }
  }
  else if( check( jx, jy, 226, 130 ) )
  {
    command( J_RIGHT, 2000 );
  }
  else if( check( jx, jy, 54, 59 ) )
  {
    command( J_BOTTOM_LEFT, 2000 );
  }
  else if( check( jx, jy, 126, 30 ) )
  {
    command( J_BOTTOM, 2000 );
  }
  else if( check( jx, jy, 200, 57 ) )
  {
    command( J_BOTTOM_RIGHT, 2000 );
  }
  else
  {
    //print_data( jx, jy, ax, ay,az, bz, bc ); // for debug purposes only leave it here
  }
}

// checks if the joystick position is in the given position with DELTA tolerance
int check( int sx, int sy, int tx, int ty ) {
  return( abs( sx - tx ) < DELTA ) && ( abs( sy - ty ) < DELTA );
}

// debug output of nunchuck's data
void print_data( int jx, int jy, int ax, int ay, int az, int bz, int bc ) 
{
  Serial.print( jx, DEC );
  Serial.print( "\t" );
  Serial.print( jy, DEC );
  Serial.print( "\t" );
  Serial.print( ax, DEC );
  Serial.print( "\t" );
  Serial.print( ay, DEC );
  Serial.print( "\t" );
  Serial.print( az, DEC );
  Serial.print( "\t" );
  Serial.print( bc, DEC );
  Serial.print( "\t" );
  Serial.print( bz, DEC );
  Serial.print( "\t" );
  Serial.print( "\r\n" );
}

// decrypts nunchuck data, famous "formula", I didn't find why it is working
char decrypt( char x )
{
  return( x ^ 0x17 ) + 0x17;
}
